
# parsetab.py
# This file is automatically generated. Do not edit.
# pylint: disable=W,C,R
_tabversion = '3.10'

_lr_method = 'LALR'

_lr_signature = 'LEFT NAME NUMBER RIGHT SEPARATORstatement : typetype : NAMEtype : NAME LEFT list RIGHTlist : arglist : list SEPARATOR argarg : NUMBERarg : type'

_lr_action_items = {'NAME':([0,4,10,],[3,3,3,]),'$end':([1,2,3,9,],[0,-1,-2,-3,]),'RIGHT':([3,5,6,7,8,9,11,],[-2,9,-4,-6,-7,-3,-5,]),'SEPARATOR':([3,5,6,7,8,9,11,],[-2,10,-4,-6,-7,-3,-5,]),'LEFT':([3,],[4,]),'NUMBER':([4,10,],[7,7,]),}

_lr_action = {}
for _k, _v in _lr_action_items.items():
   for _x,_y in zip(_v[0],_v[1]):
      if not _x in _lr_action:  _lr_action[_x] = {}
      _lr_action[_x][_k] = _y
del _lr_action_items

_lr_goto_items = {'statement':([0,],[1,]),'type':([0,4,10,],[2,8,8,]),'list':([4,],[5,]),'arg':([4,10,],[6,11,]),}

_lr_goto = {}
for _k, _v in _lr_goto_items.items():
   for _x, _y in zip(_v[0], _v[1]):
       if not _x in _lr_goto: _lr_goto[_x] = {}
       _lr_goto[_x][_k] = _y
del _lr_goto_items
_lr_productions = [
  ("S' -> statement","S'",1,None,None,None),
  ('statement -> type','statement',1,'p_statement_type','convert_type.py',146),
  ('type -> NAME','type',1,'p_type_name','convert_type.py',150),
  ('type -> NAME LEFT list RIGHT','type',4,'p_type_name_list','convert_type.py',162),
  ('list -> arg','list',1,'p_list_arg','convert_type.py',173),
  ('list -> list SEPARATOR arg','list',3,'p_list_list_arg','convert_type.py',177),
  ('arg -> NUMBER','arg',1,'p_arg_NUMBER','convert_type.py',184),
  ('arg -> type','arg',1,'p_arg_type','convert_type.py',189),
]
